UNIT SGIFImg;

{ Freeware GIF image Delphi-component for both 16 and 32 bit
  made by Endre I. Simay, Hungary, 1998.

  It's made by implementing GIF Decoding procedures
  for  Borland/Turbo Pascal  7.0
  Written  by  Tristan Tarrant,  1994
  and some tricks were got from Richard Shotbolt's
  XGif Image component.

  ( Supports the most GIF87a/GIF89a images )

  This is an enhance component as a descendant from
  TUniBitmap (from same author) and register the GIF file format,
  so You Can load a Gif from a Resource (*.RES) file
  as RCDATA

Left Open :
  - Also store the format via SaveToStream override
    so the GIF format can be used in blob formats
    also !
  - GIF 89a format might be not work propably, especially the animated ones
  - support of new PNG format

Before using the GIF format in any commercial
application be sure you know the legal issuees for
this format !!!

}

INTERFACE

USES
  Forms, WinTypes, WinProcs, Graphics, SysUtils, Classes,
  UniBmp;

CONST
  MAX_WIDTH = 2*4000;    { arbitrary - maximum width (in bytes) of a GIF image }
  MAX_BLOCK = 2*4096;
TYPE
{$IFDEF WIN32}
  Integer16 = Smallint;
{$ELSE}
  Integer16 = Integer;
{$ENDIF}
TYPE
  block_ARRAY = ARRAY [0..MAX_BLOCK] OF Byte;
  line_ARRAY = PACKED ARRAY [0..MAX_WIDTH] OF Byte;
TYPE
  TByteArray = ARRAY [0..0] OF Byte; {Array of bytes, useful for typecasting}
  TIntArray  = ARRAY [0..0] OF Integer16; {Array of Integer16s, useful for typecasting}
  TWordArray = ARRAY [0..0] OF Integer16; {Array of words, useful for typecasting}
  TCharArray = ARRAY [0..0] OF Char; {Array of chars, useful for typecasting}
CONST
  Red = 0;
  Green = 1;
  Blue = 2;

TYPE
  PBmLine         = ^TBmpLinesStruct;
  TBmpLinesStruct = PACKED RECORD
                             LineData  : Line_ARRAY;
                             LineNo    : LongInt;
                           END;

CONST
{ Error constants used in GIF decoder }
  GoodRead      = 0; { No errors encountered during encoding/decoding }
  BadFile       = 1; { Physical problem with the media}
  BadRead       = 2; { Could not read/interpret part of the file }
  UnexpectedEOF = 3; { File too short during decoding}
  BadCode       = 4; { Code encountered during decoding was not expected}
  BadFirstCode  = 5; { The first code was invalid}
  NoFile        = 6; { Could not open the file for read/write}
  BadSymbolSize = 7; { Number of bits not supported}
  NoCode        = - 1;

TYPE
  GifHeader =  RECORD
                 sig : ARRAY [1..6] OF Char;
                 screenwidth, screendepth : Integer16;
                 Flags, background, aspect : Byte;
               END;

  ImageBlock = RECORD
                 Left, Top, width, depth : Word;
                 Flags : Byte;
               END;

  FileInfo =  RECORD
                width, depth, bits,
                Flags, background : Integer16;
                palette : ARRAY [1..768] OF Byte;
              END;

  ControlBlock =  RECORD
                    blocksize, Flags : Byte;
                    Delay : Word;
                    transparentcolour, terminator : Byte;
                  END;

  PlainText =  RECORD
                 blocksize : Byte;
                 Left, Top, gridwidth, gridheight : Word;
                 cellwidth, cellheight, forecolour, backcolour : Byte;
               END;

  GifApp =  RECORD
                   blocksize : Byte;
                   applstring : ARRAY [1..8] OF Char;
                   authentication : ARRAY [1..3] OF Char;
                 END;

TYPE
  EGIFException = CLASS (Exception)
                  END;

  TSGIF = CLASS (TPersistent)
            PRIVATE
            GifPalette : ARRAY [0..255, Red..Blue] OF Byte;
            { GifPalette is an array which contains the palette of the last loaded
              GIF file}
            FStream    : TStream;            { the file stream for the Gif file }
            FOStream   : TMemoryStream;      { the file output stream for the image }
            BmIHeader  : TBitmapInfoHeader;  { File Header for bitmap file}
            BmFHeader  : TBitmapFileHeader;  { File Header for bitmap file}
            ImageLines : TList;              { Image data}
            LineBuffer : line_ARRAY;         { array for buffer line output }
            UseLocalColors: Boolean;
            FHEight,
            FWidth,
            CurrentY   : LongInt;            { current screen-line location }
            PUBLIC
            HasTranparent:Boolean;
            FTransparentCol:TColor;
            CONSTRUCTOR Create;
            DESTRUCTOR  Destroy; override;
            PROCEDURE   LoadFromStream (Stream: TStream);
            PROCEDURE   SaveToStream (Stream: TStream);
            PROCEDURE   CreateLine;
            FUNCTION    LoadGif: Integer16;
            {The palette of the picture is stored in the
            global variable GifPalette}
            FUNCTION    GifError ( ErrorCode : Integer16 ) : STRING;
                       { This function converts an error code returned by SaveGIF into a string}
            FUNCTION UnpackImage ( bits : Integer16; VAR fi : FileInfo ) : Integer16;
            PROCEDURE SkipExtension;
            FUNCTION UnpackGIF : Integer16;
          END;

  TSGIFBitmap = CLASS (TUniBitmap)
                  Private
                  FTransColor:TColor;
                  Loaded:Boolean;
                  Function GetTransparent:TColor;
                  PUBLIC
                  Constructor Create; override;
                  PROCEDURE LoadFromStream (Stream: TStream); override;
                  property TransparentColor:TColor read GetTransparent;
                END;

IMPLEMENTATION

Constructor TSGIFBitmap.Create;
begin
 Inherited Create;
 Loaded:=False;
end;

Function TSGIFBitmap.GetTransparent:TColor;
var
 bmp:TBitmap;
begin
if Loaded then
  result:=FTransColor
 else
  begin
    bmp:=TBitmap.Create;
    try
     Bmp.Assign(Self);
     result:=Bmp.TransparentColor;
    finally
     Bmp.free;
    end;
   end;
end;

PROCEDURE TSGIFBitmap. LoadFromStream (Stream: TStream);
VAR
  DBmp: TBitmap;
  SP: TSGIF;
  MS: TMemoryStream;
BEGIN
  DBmp := TBitmap. Create;
  SP := TSGIF. Create;
  MS := TMemoryStream. Create;
  try
  SP. LoadFromStream (Stream);
  SP. SaveToStream (MS);
  MS. Seek (0, 0);
  DBmp. LoadFromStream (MS);
  Assign (DBmp);
  Loaded:=True;
  if SP.HasTranparent Then
   FTransColor:=SP.FTransparentCol
   else FTransColor:=DBmp.TransParentColor;
  finally
  DBmp. Free;
  SP. Free;
  MS. Free;
 END;
END;

CONSTRUCTOR TSGIF. Create;
BEGIN
  INHERITED Create;
  FOStream := TMemoryStream. Create;
  FStream := NIL;
  ImageLines := TList. Create;
  HasTranparent:=False;
  FTransparentCol:=0;
END;

DESTRUCTOR  TSGIF. Destroy;
BEGIN
  FOStream. Free;
  ImageLines. Free;
  INHERITED Destroy;
END;

PROCEDURE TSGIF. LoadFromStream (Stream: TStream);
VAR I: Integer16;
BEGIN
  FStream := Stream;
  FStream. Seek (0, 0);
  I := LoadGif;
  CurrentY := 0;
  IF I <> 0 THEN
   raise EGIFException. Create (GifError (i) );
END;

PROCEDURE TSGIF. SaveToStream (Stream: TStream);
VAR
  i: LongInt;
  Line: LongInt;
  CH: Char;
  p: PBmLine;
  X: LongInt;
CONST  j: Byte = 0;
BEGIN
  WITH BmFHeader DO BEGIN
    bfType := (ORD('M') SHL 8)+ORD('B');
    bfSize := (3 * 255) + SizeOf (TBitmapFileHeader) +  {Color map info}
              SizeOf (TBitmapInfoHeader) + (fHeight * fWidth);
    bfReserved1 := 0; {not currently used}
    bfReserved2 := 0; {not currently used}
    bfOffBits := (4 * 256) + SizeOf (TBitmapFileHeader) +
                 SizeOf (TBitmapInfoHeader);
  END;
  WITH Stream DO
  BEGIN
    Write (BmFHeader, SizeOf (BmFHeader) );
    WITH BmIHeader DO
    BEGIN
      biSize           := SizeOf (TBitmapInfoHeader);
      biWidth          := FWidth;
      biHeight         := FHeight;
      biPlanes         := 1;            {Arcane and rarely used}
      biBitCount       := 8;            {Hmmm Should this be hardcoded ?}
      biCompression    := BI_RGB;       {Sorry Did not implement compression in this version}
      biSizeImage      := 0;            {Valid since we are not compressing the image}
      biXPelsPerMeter  := 143;           {Rarely used very arcane field}
      biYPelsPerMeter  := 143;           {Ditto}
      biClrUsed        := 0;            {all colors are used}
      biClrImportant   := 0;            {all colors are important}
    END;
    Write (BmIHeader, SizeOf (BmIHeader) );
  BEGIN
    FOR i := 0 TO 255 DO
    BEGIN
      Write (GifPalette [i, Blue], 1);
      Write (GifPalette [i, Green], 1);
      Write (GifPalette [i, Red], 1);
      Write (j, 1); {Bogus palete entry required by windows}
    END;
  END;

  Line := FHeight;
  {Write out File lines in reverse order}
  WHILE Line >= 0 DO
  BEGIN
    {Go through the line list in reverse order looking for the
    current Line. Use reverse order since non interlaced gifs are
    stored top to bottom.  Bmp file need to be written bottom to
    top}
    FOR i := (ImageLines. Count - 1) DOWNTO 0  DO
    BEGIN
      p := ImageLines. Items [i];
      IF p^. LineNo = Line THEN
      BEGIN
        X := FWidth{ImageDescriptor.ImageWidth} ;
        Write (p^. LineData, X);
        CH := Chr (0);
        WHILE (X AND 3) <> 0 DO { Pad up to 4-byte boundary with zeroes }
        BEGIN
          Inc (X);
          Write (CH, 1);
        END;
        Break;
      END;
    END;
    Dec (Line);
  END;
  Position := 0; { reset mewmory stream}
END;


END;

CONST
  TableSize = 5003;
{ These values will be masked with the codes output from the
  decoder to remove spurious bits }
  CodeMask : ARRAY [1..13] OF Word =
  ( $0000,
  $0001, $0003,
  $0007, $000F,
  $001F, $003F,
  $007F, $00FF,
  $01FF, $03FF,
  $07FF, $0FFF );
  LargestCode = 4095;

PROCEDURE TSGIF. CreateLine;
VAR
  p: PBmLine;
BEGIN
  Forms. Application. ProcessMessages;
{Create a new bmp line}
  New (p);
  p^. LineData := LineBuffer;
  p^. LineNo := CurrentY;
  ImageLines. Add (p);
{Prepare for the next line}
  Inc (CurrentY);
END;

FUNCTION TSGIF. UnpackImage ( bits : Integer16; VAR fi : FileInfo ) : Integer16;
VAR
  bits2, codesize, codesize2, nextcode, thiscode,
  oldtoken, currentcode, oldcode, bitsleft, blocksize,
  Line, pass, byt, p, q, u : Integer16;
  b : ARRAY [0..255] OF Byte;
  firstcodestack, lastcodestack : ^TByteArray;
  codestack : ^TIntArray;
CONST
  wordmasktable : ARRAY [0..15] OF Word =
  ( $0000, $0001, $0003, $0007, $000F, $001F,
  $003F, $007F, $00FF, $01FF, $03FF, $07FF,
  $0FFF, $1FFF, $3FFF, $7FFF );
  inctable : ARRAY [0..4] OF Integer16 = ( 8, 8, 4, 2, 0 );
  starttable : ARRAY [0..4] OF Integer16 = ( 0, 4, 2, 1, 0 );
BEGIN
  FWidth := fi. width;
  FHEight := fi. depth;
  pass := 0;
  Line := 0;
  byt := 0;
  p := 0;
  q := 0;
  blocksize := 0;
  FillChar ( b, 256, 0 );
  bitsleft := 8;
  IF ( bits < 2 ) OR ( bits > 8 ) THEN
  BEGIN
    UnpackImage := BadSymbolSize;
    Exit;
  END;
  bits2 := 1 SHL bits;
  nextcode := bits2 + 2;
  codesize := bits + 1;
  codesize2 := 1 SHL codesize;
  oldcode := NoCode;
  oldtoken := NoCode;
  GetMem ( firstcodestack, 4096 );
  GetMem ( lastcodestack, 4096 );
  GetMem ( codestack, 8192 );

  WHILE TRUE DO
  BEGIN
    IF bitsleft = 8 THEN
    BEGIN
      Inc (p);
      IF p >= q THEN
      BEGIN
        blocksize := 0;
        FStream. Read (blocksize, 1);
        IF blocksize > 0 THEN
        BEGIN
          p := 0;
          q := FStream. Read (b, blocksize);
          IF q <> blocksize THEN
          BEGIN
            FreeMem ( firstcodestack, 4096 );
            FreeMem ( lastcodestack, 4096 );
            FreeMem ( codestack, 8192 );
            UnpackImage := UnexpectedEOF;
            Exit;
          END;
        END ELSE
        BEGIN
          FreeMem ( firstcodestack, 4096 );
          FreeMem ( lastcodestack, 4096 );
          FreeMem ( codestack, 8192 );
          UnpackImage := UnexpectedEOF;
          Exit;
        END;
      END;
      bitsleft := 0;
    END;
    thiscode := b [p];
    currentcode := codesize + bitsleft;
    IF currentcode <= 8 THEN
    BEGIN
      b [p] := b [p] SHR codesize;
      bitsleft := currentcode;
    END ELSE
    BEGIN
      Inc (p);
      IF p >= q THEN
      BEGIN
        blocksize := 0;
        FStream. Read ( blocksize, 1);
        IF blocksize > 0 THEN
        BEGIN
          p := 0;
          q := FStream. Read ( b, blocksize );
          IF q <> blocksize THEN
          BEGIN
            FreeMem ( firstcodestack, 4096 );
            FreeMem ( lastcodestack, 4096 );
            FreeMem ( codestack, 8192 );
            UnpackImage := UnexpectedEOF;
            Exit;
          END;
        END ELSE
        BEGIN
          FreeMem ( firstcodestack, 4096 );
          FreeMem ( lastcodestack, 4096 );
          FreeMem ( codestack, 8192 );
          UnpackImage := UnexpectedEOF;
          Exit;
        END;
      END;
      thiscode := thiscode OR ( b [p] SHL (8 - bitsleft) );
      IF currentcode <= 16 THEN
      BEGIN
        bitsleft := currentcode - 8;
        b [p] := b [p] SHR bitsleft;
      END ELSE
      BEGIN
        Inc (p);
        IF p >= q THEN
        BEGIN
          blocksize := 0;
          FStream. Read ( blocksize, 1);
          IF blocksize > 0 THEN
          BEGIN
            p := 0;
            q := FStream. Read ( b, blocksize );
            IF q <> blocksize THEN
            BEGIN
              FreeMem ( firstcodestack, 4096 );
              FreeMem ( lastcodestack, 4096 );
              FreeMem ( codestack, 8192 );
              UnpackImage := UnexpectedEOF;
              Exit;
            END;
          END ELSE
          BEGIN
            FreeMem ( firstcodestack, 4096 );
            FreeMem ( lastcodestack, 4096 );
            FreeMem ( codestack, 8192 );
            UnpackImage := UnexpectedEOF;
            Exit;
          END;
        END;
        thiscode := thiscode OR ( b [p] SHL (16 - bitsleft) );
        bitsleft := currentcode - 16;
        b [p] := b [p] SHR bitsleft;
      END;
    END;
    thiscode := thiscode AND wordmasktable [codesize];
    currentcode := thiscode;
    IF thiscode = bits2 + 1 THEN Break;
    IF thiscode > nextcode THEN
    BEGIN
      FreeMem ( firstcodestack, 4096 );
      FreeMem ( lastcodestack, 4096 );
      FreeMem ( codestack, 8192 );
      UnpackImage := BadCode;
      Exit;
    END;
    IF thiscode = bits2 THEN
    BEGIN
      nextcode := bits2 + 2;
      codesize := bits + 1;
      codesize2 := 1 SHL codesize;
      oldtoken := NoCode;
      OldCode := NoCode;
      Continue;
    END;
    u := 0;
    IF thiscode = nextcode THEN
    BEGIN
      IF oldcode = NoCode THEN
      BEGIN
        FreeMem ( firstcodestack, 4096 );
        FreeMem ( lastcodestack, 4096 );
        FreeMem ( codestack, 8192 );
        UnpackImage := BadFirstCode;
        Exit;
      END;
      firstcodestack^ [u] := oldtoken;
      Inc ( u );
      thiscode := oldcode;
    END;
    WHILE thiscode >= bits2 DO
    BEGIN
      firstcodestack^ [u] := lastcodestack^ [thiscode];
      Inc ( u );
      thiscode := codestack^ [thiscode];
    END;
    oldtoken := thiscode;
    WHILE TRUE DO
    BEGIN
      linebuffer [byt] := thiscode;
      Inc ( byt );
      IF byt >= fi. width THEN
      BEGIN
        currenty := Line;
        CreateLine;
        byt := 0;
        IF fi. Flags AND $40 = $40 THEN
        BEGIN
          Line := Line + inctable [pass];
          IF Line >= fi. depth THEN
          BEGIN
            Inc (pass);
            Line := starttable [pass];
          END;
        END ELSE Inc (Line);
      END;
      IF u <= 0 THEN Break;
      Dec ( u );
      thiscode := firstcodestack^ [u];
    END;
    IF (nextcode < 4096) AND (oldcode <> NoCode) THEN
    BEGIN
      codestack^ [nextcode] := oldcode;
      lastcodestack^ [nextcode] := oldtoken;
      Inc ( nextcode );
      IF (nextcode >= codesize2) AND (codesize < 12) THEN
      BEGIN
        Inc ( codesize );
        codesize2 := 1 SHL codesize;
      END;
    END;
    oldcode := currentcode;
  END;
  FreeMem ( firstcodestack, 4096 );
  FreeMem ( lastcodestack, 4096 );
  FreeMem ( codestack, 8192 );
  UnpackImage := GoodRead;
END; { UnpackImage }

PROCEDURE TSGIF. SkipExtension;
VAR
  pt : PlainText;
  cb : ControlBlock;
  ap : GifApp;
  i : Integer16;
  a, n, c : Byte;
  r : Word;
BEGIN
  FStream. Read ( c, 1 );
  CASE c OF
    $01 :
         BEGIN
           FStream. Read ( pt, 13{SizeOf ( PlainText )} );
           FStream. Read (  n, 1 );
           WHILE n > 0 DO
           BEGIN
             FOR i := 0 TO n - 1 DO
               FStream. Read ( a, 1 );
             FStream. Read	 ( n, 1 );
           END;
         END;
    $F9 :BEGIN
          FStream. Read ( cb, 6{SizeOf ( ControlBlock ) });
          HasTranparent:=True;
          FTransparentCol:=
          RGB( GifPalette[cb.TransparentColour,Red],
               GifPalette[cb.TransparentColour,Green],
               GifPalette[cb.TransparentColour,Blue]);
         END;
    $FE :
         BEGIN
           FStream. Read ( n, 1 );
           WHILE n > 0 DO
           BEGIN
             FOR i := 0 TO n - 1 DO
               FStream. Read ( a, 1 );
             FStream. Read (  n, 1 );
           END;
         END;
    $FF :
         BEGIN
           FStream. Read ( ap, 12{SizeOf ( GifApp )} );
           FStream. Read ( n, 1 );
           WHILE n > 0 DO
           BEGIN
             FOR i := 0 TO n - 1 DO
               FStream. Read ( a, 1 );
             FStream. Read ( n, 1 );
           END;
         END;
    ELSE
    BEGIN
      FStream. Read ( n, 1 );
      FOR i := 0 TO n - 1 DO
        FStream. Read ( a, 1 );
    END;
  END;
END; { SkipExtension }

FUNCTION TSGIF. UnpackGIF: Integer16;
VAR
  gh : GifHeader;
  iblk : ImageBlock;
  t : LongInt;
  b, c : Integer16;
  r : Word;
  CH : Char;
  fi : FileInfo;
BEGIN
  r := FStream. Read (gh,13{ SizeOf (GifHeader)} );
  IF ( gh. sig [1] + gh. sig [2] + gh. sig [3] <> 'GIF' ) OR ( r <> 13{SizeOf (GifHeader)} ) THEN
  BEGIN
    UnpackGIF := BadFile;
    Exit;
  END;
  fi. width := gh. screenwidth;
  fi. depth := gh. screendepth;
  fi. bits := gh. Flags AND $07 + 1;
  fi. background := gh. background;
  UseLocalColors := FALSE;
  IF ( gh. Flags AND $80 ) = $80 THEN
  BEGIN
    UseLocalColors := TRUE;
    c := 3 * ( 1 SHL fi. bits );
    r := FStream. Read (fi. palette, c);
    IF r <> c THEN
    BEGIN
      UnpackGIF := BadRead;
      Exit;
    END;
    FOR b := 0 TO 255 DO
    BEGIN
      GIFPalette [b, Red] := fi. palette [b * 3 + 1] ;
      GIFPalette [b, Green] := fi. palette [b * 3 + 2];
      GIFPalette [b, Blue] := fi. palette [b * 3 + 3];
    END;

  END;
  FStream. Read ( CH, 1 );
  WHILE ( CH = ',' ) OR ( CH = '!' ) OR ( CH = #0 ) DO
  BEGIN
    CASE CH OF
      ',' :
            BEGIN
              r := FStream. Read ( iblk, 9{SizeOf (ImageBlock)} );
              IF r <>  9 {SizeOf (ImageBlock)} THEN
              BEGIN
                UnpackGIF := BadRead;
                Exit;
              END;
              fi. width := iblk. width;
              fi. depth := iblk. depth;
              fHeight := iblk. depth ;
              UseLocalColors := FALSE;
              IF ( iblk. Flags AND $80 ) = $80 THEN
              BEGIN
                UseLocalColors := TRUE;
                b := 3 * (1 SHL (iblk. Flags AND $07 + 1) );
                r := FStream. Read (fi. palette, b );
                IF r <> b THEN
                BEGIN
                  UnpackGIF := BadRead;
                  Exit;
                END;
                FOR b := 0 TO 255 DO
                BEGIN
                  GIFPalette [b, Red] := fi. palette [b * 3 + 1];
                  GIFPalette [b, Green] := fi. palette [b * 3 + 2];
                  GIFPalette [b, Blue] := fi. palette [b * 3 + 3];
                END;
              END;
              IF FStream. Position = FStream. Size THEN
              BEGIN
                UnpackGIF := BadFile;
                Exit;
              END;
              c := 0;
              FStream. Read ( c, 1 );
              fi. Flags := iblk. Flags;
              t := UnpackImage (c, fi );
              UnpackGif := t;
              Exit;
            END;
      '!' :
            BEGIN SkipExtension;
              FStream. Read ( CH, 1 );
            END;
    END;
  END;
END; { UnpackGIF }

FUNCTION TSGIF. LoadGif;
BEGIN
  LoadGif := UnpackGif
END; { LoadGif }

FUNCTION TSGIF. GifError;
BEGIN
  CASE ErrorCode OF
    GoodRead : GifError := 'Ok';
    BadFile  : GifError := 'Bad File';
    BadRead  : GifError := 'Bad Read';
    UnexpectedEOF : GifError := 'Unexpected End';
    BadCode       : GifError := 'Bad LZW Code';
    BadFirstCode  : GifError := 'Bad First Code';
    BadSymbolSize : GifError := 'Bad Symbol Size';
    NoFile        : GifError := 'File Not Found';
    ELSE GifError := 'Unknown';
  END;
END; { GifError }

initialization
  {
    register the TSGifBitmap as a new graphic file format
    now all the TPicture storage stuff can access our new
    GIF graphic format !
  }
  TPicture.RegisterFileFormat('gif','GIF-Format', TSGifBitmap);
END.
